/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

#include "simple.h"
#include "token.h"

void
bi_identity(token_list *expansion,token tk,int argc,token_list *argv)
{
  copylist(expansion,argv);
}

void
bi_void(token_list *expansion,token tk,int argc,token_list *argv)
{
  makeemptylist(expansion);
}

void
bi_output(token_list *expansion,token tk,int argc,token_list *argv)
{
  int i;
  for (i=0;i<argv->size;i++)
    outputtoken(argv[0].l[i]);
}

void
bi_define(token_list *expansion,token tk,int argc,token_list *argv)
{
  token todef;
  if (argc<2)
    fatal(0,"@def@ expects two arguments.");
  if (argv[0].size>1)
    fatal(0,"Only single tokens can be defined.");
  todef=argv[0].l[0];
  if (todef<FIRST_USER_COMMAND)
    fatal(0,"Trying to define something you shouldn't.");
  defineuser(argv+1,todef);
}

void
bi_include(token_list *expansion,token tk,int argc,token_list *argv)
{
  char *name;
  name=tokenlisttostring(argv);
  pushinputfile(name);
  free(name);
}

void
bi_tokenize(token_list *expansion,token tk,int argc,token_list *argv)
{
  char *name;
  name=tokenlisttostring(argv);
  appendtoken(expansion,identifycommand(name));
  free(name);
}

void
bi_quote(token_list *expansion,token tk,int argc,token_list *argv)
{
  int i;
  ensurelength(expansion,argv[0].size*2);
  expansion->size=argv[0].size*2;
  for (i=0;i<argv[0].size;i++) {
    expansion->l[2*i]=QUOTE_NEXT;
    expansion->l[2*i+1]=argv[0].l[i];
  }
}

void
bi_if(token_list *expansion,token tk,int argc,token_list *argv)
{
  int i;
  if (argc%3==2)
    fatal(0,"@if@ can't deal with %d parameters.",argc);
  for (i=0;3*i+2<argc;i++) {
    if (comparelist(argv+3*i,argv+(3*i+1))) {
      copylist(expansion,argv+(3*i+2));
      return;
    }
  }
  if (3*i<argc) {
    copylist(expansion,argv+3*i);
  } /* Otherwise expansion is empty */
}

void
bi_head(token_list *expansion,token tk,int argc,token_list *argv)
{
  if (argv->size==0)
    fatal(0,"@head@ of an empty list is invalid.");
  appendtoken(expansion,argv->l[0]);
}

void
bi_tail(token_list *expansion,token tk,int argc,token_list *argv)
{
  if (argv->size==0)
    fatal(0,"@tail@ of an empty list is invalid.");
  copylist(expansion,argv);
  removefirst(expansion);
}

void
bi_ahead(token_list *expansion,token tk,int argc,token_list *argv)
{
  if (argv->size==0)
    fatal(0,"@ahead@ of an empty list is invalid.");
  appendtoken(expansion,argv->l[argv->size-1]);
}

void
bi_atail(token_list *expansion,token tk,int argc,token_list *argv)
{
  if (argv->size==0)
    fatal(0,"@atail@ of an empty list is invalid.");
  copylist(expansion,argv);
  removelast(expansion);
}

static struct {
  token who;
  void (* what)(token_list *,token,int,token_list *);
} builtin_table[] = {
  {BI_IDENTITY, bi_identity},
  {BI_VOID, bi_void},
  {BI_OUTPUT, bi_output},
  {BI_DEFINE, bi_define},
  {BI_INCLUDE, bi_include},
  {BI_TOKENIZE, bi_tokenize},
  {BI_QUOTE, bi_quote},
  {BI_IF, bi_if},
  {BI_HEAD, bi_head},
  {BI_TAIL, bi_tail},
  {BI_AHEAD, bi_ahead},
  {BI_ATAIL, bi_atail},
  {0, NULL}
};

void
callbuiltin(token_list *expansion,token tk,int argc,token_list *argv)
{
  int i;
  for (i=0;builtin_table[i].who;i++) {
    if (tk==builtin_table[i].who) {
      builtin_table[i].what(expansion,tk,argc,argv);
      return;
    }
  }
  fatal(1,"Unresolved builtin!");
}
